@@ -1,8 +1,5 @@
 import {Parser} from "./parse"
 
-// FIXME profile adding a per-Tree TreeNode cache, validating it by
-// parent pointer
-
 /// The default maximum length of a `TreeBuffer` node.
 export const DefaultBufferLength = 1024
 
@@ -116,6 +113,11 @@ export class MountedTree {
     /// The parser used to create this subtree.
     readonly parser: Parser
   ) {}
+
+  /// @internal
+  static get(tree: Tree | null): MountedTree | null {
+    return tree && tree.props && tree.props[NodeProp.mounted.id]
+  }
 }
 
 /// Type returned by [`NodeProp.add`](#common.NodeProp.add). Describes
@@ -337,7 +339,7 @@ export class Tree {
 
   /// @internal
   toString(): string {
-    let mounted = this.prop(NodeProp.mounted)
+    let mounted = MountedTree.get(this)
     if (mounted && !mounted.overlay) return mounted.tree.toString()
     let children = ""
     for (let ch of this.children) {
@@ -358,14 +360,14 @@ export class Tree {
   /// Get a [tree cursor](#common.TreeCursor) positioned at the top of
   /// the tree. Mode can be used to [control](#common.IterMode) which
   /// nodes the cursor visits.
-  cursor(mode: IterMode = 0) {
+  cursor(mode: IterMode = 0 as IterMode) {
     return new TreeCursor(this.topNode as TreeNode, mode)
   }
 
   /// Get a [tree cursor](#common.TreeCursor) pointing into this tree
   /// at the given position and side (see
   /// [`moveTo`](#common.TreeCursor.moveTo).
-  cursorAt(pos: number, side: -1 | 0 | 1 = 0, mode: IterMode = 0): TreeCursor {
+  cursorAt(pos: number, side: -1 | 0 | 1 = 0, mode: IterMode = 0 as IterMode): TreeCursor {
     let scope = CachedNode.get(this) || this.topNode
     let cursor = new TreeCursor(scope as TreeNode | BufferNode)
     cursor.moveTo(pos, side)
@@ -405,6 +407,15 @@ export class Tree {
     return node
   }
 
+  /// In some situations, it can be useful to iterate through all
+  /// nodes around a position, including those in overlays that don't
+  /// directly cover the position. This method gives you an iterator
+  /// that will produce all nodes, from small to big, around the given
+  /// position.
+  resolveStack(pos: number, side: -1 | 0 | 1 = 0): NodeIterator {
+    return stackIterator(this, pos, side)
+  }
+
   /// Iterate over the tree and its children, calling `enter` for any
   /// node that touches the `from`/`to` region (if given) before
   /// running over such a node's children, and `leave` (if given) when
@@ -467,6 +478,9 @@ export class Tree {
   static build(data: BuildData) { return buildTree(data) }
 }
 
+/// Represents a sequence of nodes.
+export type NodeIterator = {node: SyntaxNode, next: NodeIterator | null}
+
 type BuildData = {
   /// The buffer or buffer cursor to read the node data from.
   ///
@@ -640,7 +654,9 @@ export interface SyntaxNodeRef {
   /// Retrieve a stable [syntax node](#common.SyntaxNode) at this
   /// position.
   readonly node: SyntaxNode
-  /// Test whether the node matches a given context.
+  /// Test whether the node matches a given context—a sequence of
+  /// direct parent nodes. Empty strings in the context array act as
+  /// wildcards, other strings must match the ancestor node's name.
   matchContext(context: readonly string[]): boolean
 }
 
@@ -704,11 +720,6 @@ export interface SyntaxNode extends SyntaxNodeRef {
   /// Like [`getChild`](#common.SyntaxNode.getChild), but return all
   /// matching children, not just the first.
   getChildren(type: string | number, before?: string | number | null, after?: string | number | null): SyntaxNode[]
-
-  /// Test whether the node matches a given context—a sequence of
-  /// direct parent nodes. Empty strings in the context array act as
-  /// wildcards, other strings must match the ancestor node's name.
-  matchContext(context: readonly string[]): boolean
 }
 
 const enum Side {
@@ -731,21 +742,6 @@ function checkSide(side: Side, pos: number, from: number, to: number) {
   }
 }
 
-function enterUnfinishedNodesBefore(node: SyntaxNode, pos: number) {
-  let scan = node.childBefore(pos)
-  while (scan) {
-    let last = scan.lastChild
-    if (!last || last.to != scan.to) break
-    if (last.type.isError && last.from == last.to) {
-      node = scan
-      scan = last.prevSibling
-    } else {
-      scan = last
-    }
-  }
-  return node
-}
-
 function resolveNode(node: SyntaxNode, pos: number, side: -1 | 0 | 1, overlays: boolean): SyntaxNode {
   // Move up to a node that actually holds the position, if possible
   while (node.from == node.to ||
@@ -768,12 +764,71 @@ function resolveNode(node: SyntaxNode, pos: number, side: -1 | 0 | 1, overlays:
   }
 }
 
-export class TreeNode implements SyntaxNode {
+abstract class BaseNode implements SyntaxNode {
+  abstract from: number
+  abstract to: number
+  abstract type: NodeType
+  abstract name: string
+  abstract tree: Tree | null
+  abstract parent: SyntaxNode | null
+  abstract firstChild: SyntaxNode | null
+  abstract lastChild: SyntaxNode | null
+  abstract childAfter(pos: number): SyntaxNode | null
+  abstract childBefore(pos: number): SyntaxNode | null
+  abstract enter(pos: number, side: -1 | 0 | 1, mode?: IterMode): SyntaxNode | null
+  abstract nextSibling: SyntaxNode | null
+  abstract prevSibling: SyntaxNode | null
+  abstract toTree(): Tree
+
+  cursor(mode: IterMode = 0 as IterMode) { return new TreeCursor(this as any, mode) }
+
+  getChild(type: string | number, before: string | number | null = null, after: string | number | null = null) {
+    let r = getChildren(this, type, before, after)
+    return r.length ? r[0] : null
+  }
+
+  getChildren(type: string | number, before: string | number | null = null, after: string | number | null = null): SyntaxNode[] {
+    return getChildren(this, type, before, after)
+  }
+
+  resolve(pos: number, side: -1 | 0 | 1 = 0): SyntaxNode {
+    return resolveNode(this, pos, side, false)
+  }
+
+  resolveInner(pos: number, side: -1 | 0 | 1 = 0): SyntaxNode {
+    return resolveNode(this, pos, side, true)
+  }
+
+  matchContext(context: readonly string[]): boolean {
+    return matchNodeContext(this, context)
+  }
+
+  enterUnfinishedNodesBefore(pos: number) {
+    let scan = this.childBefore(pos), node: SyntaxNode = this
+    while (scan) {
+      let last = scan.lastChild
+      if (!last || last.to != scan.to) break
+      if (last.type.isError && last.from == last.to) {
+        node = scan
+        scan = last.prevSibling
+      } else {
+        scan = last
+      }
+    }
+    return node
+  }
+
+  get node() { return this }
+
+  get next() { return this.parent }
+}
+
+export class TreeNode extends BaseNode implements SyntaxNode {
   constructor(readonly _tree: Tree,
               readonly from: number,
               // Index in parent node, set to -1 if the node is not a direct child of _parent.node (overlay)
               readonly index: number,
-              readonly _parent: TreeNode | null) {}
+              readonly _parent: TreeNode | null) { super() }
 
   get type() { return this._tree.type }
 
@@ -781,7 +836,7 @@ export class TreeNode implements SyntaxNode {
 
   get to() { return this.from + this._tree.length }
 
-  nextChild(i: number, dir: 1 | -1, pos: number, side: Side, mode: IterMode = 0): TreeNode | BufferNode | null {
+  nextChild(i: number, dir: 1 | -1, pos: number, side: Side, mode: IterMode = 0 as IterMode): TreeNode | BufferNode | null {
     for (let parent: TreeNode = this;;) {
       for (let {children, positions} = parent._tree, e = dir > 0 ? children.length : -1; i != e; i += dir) {
         let next = children[i], start = positions[i] + parent.from
@@ -792,8 +847,7 @@ export class TreeNode implements SyntaxNode {
           if (index > -1) return new BufferNode(new BufferContext(parent, next, i, start), null, index)
         } else if ((mode & IterMode.IncludeAnonymous) || (!next.type.isAnonymous || hasChild(next))) {
           let mounted
-          if (!(mode & IterMode.IgnoreMounts) &&
-              next.props && (mounted = next.prop(NodeProp.mounted)) && !mounted.overlay)
+          if (!(mode & IterMode.IgnoreMounts) && (mounted = MountedTree.get(next)) && !mounted.overlay)
             return new TreeNode(mounted.tree, start, i, parent)
           let inner = new TreeNode(next, start, i, parent)
           return (mode & IterMode.IncludeAnonymous) || !inner.type.isAnonymous ? inner
@@ -816,7 +870,7 @@ export class TreeNode implements SyntaxNode {
 
   enter(pos: number, side: -1 | 0 | 1, mode = 0) {
     let mounted
-    if (!(mode & IterMode.IgnoreOverlays) && (mounted = this._tree.prop(NodeProp.mounted)) && mounted.overlay) {
+    if (!(mode & IterMode.IgnoreOverlays) && (mounted = MountedTree.get(this._tree)) && mounted.overlay) {
       let rPos = pos - this.from
       for (let {from, to} of mounted.overlay) {
         if ((side > 0 ? from <= rPos : from < rPos) &&
@@ -844,37 +898,12 @@ export class TreeNode implements SyntaxNode {
     return this._parent && this.index >= 0 ? this._parent.nextChild(this.index - 1, -1, 0, Side.DontCare) : null
   }
 
-  cursor(mode: IterMode = 0) { return new TreeCursor(this, mode) }
-
   get tree() { return this._tree }
 
   toTree() { return this._tree }
 
-  resolve(pos: number, side: -1 | 0 | 1 = 0) {
-    return resolveNode(this, pos, side, false)
-  }
-
-  resolveInner(pos: number, side: -1 | 0 | 1 = 0) {
-    return resolveNode(this, pos, side, true)
-  }
-
-  enterUnfinishedNodesBefore(pos: number) { return enterUnfinishedNodesBefore(this, pos) }
-
-  getChild(type: string | number, before: string | number | null = null, after: string | number | null = null) {
-    let r = getChildren(this, type, before, after)
-    return r.length ? r[0] : null
-  }
-
-  getChildren(type: string | number, before: string | number | null = null, after: string | number | null = null) {
-    return getChildren(this, type, before, after)
-  }
-
   /// @internal
   toString() { return this._tree.toString() }
-
-  get node() { return this }
-
-  matchContext(context: readonly string[]): boolean { return matchNodeContext(this, context) }
 }
 
 function getChildren(node: SyntaxNode, type: string | number, before: string | number | null, after: string | number | null): SyntaxNode[] {
@@ -906,7 +935,7 @@ class BufferContext {
               readonly start: number) {}
 }
 
-class BufferNode implements SyntaxNode {
+class BufferNode extends BaseNode {
   type: NodeType
 
   get name() { return this.type.name }
@@ -918,6 +947,7 @@ class BufferNode implements SyntaxNode {
   constructor(readonly context: BufferContext,
               readonly _parent: BufferNode | null,
               readonly index: number) {
+    super()
     this.type = context.buffer.set.types[context.buffer.buffer[index]]
   }
 
@@ -933,7 +963,7 @@ class BufferNode implements SyntaxNode {
   childAfter(pos: number) { return this.child(1, pos, Side.After) }
   childBefore(pos: number) { return this.child(-1, pos, Side.Before) }
 
-  enter(pos: number, side: -1 | 0 | 1, mode: IterMode = 0) {
+  enter(pos: number, side: -1 | 0 | 1, mode: IterMode = 0 as IterMode) {
     if (mode & IterMode.ExcludeBuffers) return null
     let {buffer} = this.context
     let index = buffer.findChild(this.index + 4, buffer.buffer[this.index + 3], side > 0 ? 1 : -1, pos - this.context.start, side)
@@ -963,8 +993,6 @@ class BufferNode implements SyntaxNode {
     return new BufferNode(this.context, this._parent, buffer.findChild(parentStart, this.index, -1, 0, Side.DontCare))
   }
 
-  cursor(mode: IterMode = 0) { return new TreeCursor(this, mode) }
-
   get tree() { return null }
 
   toTree() {
@@ -979,31 +1007,50 @@ class BufferNode implements SyntaxNode {
     return new Tree(this.type, children, positions, this.to - this.from)
   }
 
-  resolve(pos: number, side: -1 | 0 | 1 = 0) {
-    return resolveNode(this, pos, side, false)
-  }
-
-  resolveInner(pos: number, side: -1 | 0 | 1 = 0) {
-    return resolveNode(this, pos, side, true)
-  }
-
-  enterUnfinishedNodesBefore(pos: number) { return enterUnfinishedNodesBefore(this, pos) }
-
   /// @internal
   toString() { return this.context.buffer.childString(this.index) }
+}
 
-  getChild(type: string | number, before: string | number | null = null, after: string | number | null = null) {
-    let r = getChildren(this, type, before, after)
-    return r.length ? r[0] : null
+function iterStack(heads: readonly SyntaxNode[]): NodeIterator | null {
+  if (!heads.length) return null
+  if (heads.length == 1) return heads[0] as any
+
+  let pick = 0, picked = heads[0]
+  for (let i = 1; i < heads.length; i++) {
+    let node = heads[i]
+    if (node.from > picked.from || node.to < picked.to) { picked = node; pick = i }
   }
+  let next = picked instanceof TreeNode && picked.index < 0 ? null : picked.parent
+  let newHeads = heads.slice()
+  if (next) newHeads[pick] = next
+  else newHeads.splice(pick, 1)
+  return new StackIterator(newHeads, picked)
+}
 
-  getChildren(type: string | number, before: string | number | null = null, after: string | number | null = null) {
-    return getChildren(this, type, before, after)
+class StackIterator implements NodeIterator {
+  constructor(readonly heads: readonly SyntaxNode[],
+              readonly node: SyntaxNode) {}
+  get next() { return iterStack(this.heads) }
+}
+
+function stackIterator(tree: Tree, pos: number, side: -1 | 0 | 1): NodeIterator {
+  let inner = tree.resolveInner(pos, side), layers: SyntaxNode[] | null = null
+  for (let scan: TreeNode | null = inner instanceof TreeNode ? inner : (inner as BufferNode).context.parent;
+       scan; scan = scan.parent) {
+    if (scan.index < 0) { // This is an overlay root
+      let parent: TreeNode | null = scan.parent!
+      ;(layers || (layers = [inner])).push(parent.resolve(pos, side))
+      scan = parent
+    } else {
+      let mount = MountedTree.get(scan.tree)
+      // Relevant overlay branching off
+      if (mount && mount.overlay && mount.overlay[0].from <= pos && mount.overlay[mount.overlay.length - 1].to >= pos) {
+        let root = new TreeNode(mount.tree, mount.overlay[0].from + scan.from, 0, null)
+        ;(layers || (layers = [inner])).push(resolveNode(root, pos, side, false))
+      }
+    }
   }
-
-  get node() { return this }
-
-  matchContext(context: readonly string[]): boolean { return matchNodeContext(this, context) }
+  return layers ? iterStack(layers) : inner as any
 }
 
 /// A tree cursor object focuses on a given node in a syntax tree, and
